----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    16:57:30 04/08/2008 
-- Design Name: 
-- Module Name:    deser - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity deser is
    Port (
           clk : in  STD_LOGIC; -- system clock
           clk_chk : in  STD_LOGIC; -- link clock presence check, comes every 2 microseconds
           l_clk_p : in  STD_LOGIC; -- link clock in
           l_clk_n : in  STD_LOGIC; -- link clock in
           l_data_p : in  STD_LOGIC_VECTOR (3 downto 0); -- link data in
           l_data_n : in  STD_LOGIC_VECTOR (3 downto 0); -- link data in
           polarity : in  STD_LOGIC_VECTOR (3 downto 0); -- link data in polarity correction
           l_clkout : out  STD_LOGIC; -- link clock7X out
           strobe : out  STD_LOGIC; -- link clock7X strobe
           q : out  STD_LOGIC_VECTOR (27 downto 0); -- link data out
           deser_status : out  STD_LOGIC_VECTOR (11 downto 0)); -- link status
end deser;

architecture Behavioral of deser is
signal l_data : std_logic_vector(3 downto 0) := (others =>'0');
signal l_data_q : std_logic_vector(3 downto 0) := (others =>'0');
signal l_data_sr : std_logic_vector(27 downto 0) := (others =>'0');
signal strobe_p : std_logic := '0';
signal strobe_i : std_logic := '0';
signal l_clk_tgl : std_logic := '0';
signal l_clk_sync : std_logic_vector(2 downto 0) := (others =>'0');
signal l_clk_cntr : std_logic_vector(6 downto 0) := (others =>'0');
signal bound : std_logic_vector(1 downto 0) := (others =>'0');
signal err_cntr : std_logic_vector(1 downto 0) := (others =>'0');
signal rst_dcm : std_logic := '0';
signal clk0 : std_logic := '0';
signal clkfx : std_logic := '0';
signal l_clk : std_logic := '0';
signal l_clk_pq : std_logic_vector(4 downto 0) := (others =>'0');
signal l_clk_nq : std_logic_vector(2 downto 0) := (others =>'0');
signal l_clk7x : std_logic := '0';
signal locked : std_logic := '0';
signal lock_cntr : std_logic_vector(8 downto 0) := (others =>'0');
signal ps_rdy : std_logic := '0';
signal psdone : std_logic := '0';
signal psincdec : std_logic := '0';
signal psen : std_logic := '0';
signal ps_max : std_logic := '0';
signal ps_min : std_logic := '0';
signal downcnt : std_logic := '0';
signal downcnt_sync : std_logic := '0';
signal ps_cntr : std_logic_vector(3 downto 0) := (others =>'0');
signal status_i : std_logic_vector(7 downto 0) := (others =>'0');
signal phase_shift : std_logic_vector(7 downto 0) := (others =>'0');
attribute IOB : string;
attribute IOB of l_data_q : signal is "true";
begin
strobe <= strobe_p;
l_clkout <= l_clk7x;
deser_status <= not locked & status_i(2 downto 0) & phase_shift;
i_clock: ibufds generic map(DIFF_TERM => TRUE,IOSTANDARD => "LVDS_33") port map(i => l_clk_p, ib => l_clk_n, o => l_clk);
g_l_data: for i in l_data_p'range generate
  i_l_data: ibufds generic map(DIFF_TERM => TRUE,IOSTANDARD => "LVDS_33") port map(i => l_data_p(i), ib => l_data_n(i), o => l_data(i));
end generate;
i_l_clk7x : bufg port map(i => clkfx, o => l_clk7x);
i_DCM : DCM_SP
   generic map (
      CLKFX_MULTIPLY => 7, --  Can be any integer from 1 to 32
      CLKIN_PERIOD => 25.0, --  Specify period of input clock
			DESKEW_ADJUST => "SOURCE_SYNCHRONOUS",
      CLKOUT_PHASE_SHIFT => "VARIABLE") --  Specify phase shift of "NONE", "FIXED" or "VARIABLE" 
   port map (
      CLK0 => CLK0,     -- 0 degree DCM CLK ouptput
      CLKFX180 => CLKFX, -- 180 degree CLK synthesis out
      LOCKED => LOCKED, -- DCM LOCK status output
      PSDONE => PSDONE, -- Dynamic phase adjust done output
      STATUS => STATUS_i, -- 8-bit DCM status bits output
      CLKFB => CLK0,   -- DCM clock feedback
      CLKIN => l_clk,   -- Clock input (from IBUFG, BUFG or DCM)
      PSCLK => clk,   -- Dynamic phase adjust clock input
      PSEN => PSEN,     -- Dynamic phase adjust enable input
      PSINCDEC => psincdec, -- Dynamic phase adjust increment/decrement
      RST => rst_dcm        -- DCM asynchronous reset input
   );
process(clk,locked)
begin
  if(locked = '0')then
    ps_rdy <= '1';
		psen <= '0';
		psincdec <= '0';
		phase_shift <= x"80";
		ps_cntr <= x"8";
		ps_max <= '0';
		ps_min <= '0';
  elsif(clk'event and clk = '1')then
    ps_rdy <= not psen and (ps_rdy or psdone);
		if(psen = '0' and ps_rdy = '1' and (ps_cntr = x"f" or ps_cntr = x"0"))then
			psen <= '1';
			psincdec <= ps_cntr(3);
	  else
			psen <= '0';
	  end if;
		if(psen = '1')then
			if(psincdec = '1' and ps_max = '0')then
				phase_shift <= phase_shift + 1;
			elsif(psincdec = '0' and ps_min = '0')then
				phase_shift <= phase_shift - 1;
			end if;
		end if;
		if(psen = '1')then
			ps_cntr <= x"8";
		elsif(l_clk_sync(2) /= l_clk_sync(1) and downcnt_sync = '1' and ps_cntr /= x"0")then
			ps_cntr <= ps_cntr - 1;
		elsif(l_clk_sync(2) /= l_clk_sync(1) and downcnt_sync = '0' and ps_cntr /= x"f")then
			ps_cntr <= ps_cntr + 1;
		end if;
		if(phase_shift = x"ff")then
			ps_max <= '1';
		else
			ps_max <= '0';
		end if;
		if(phase_shift = x"00")then
			ps_min <= '1';
		else
			ps_min <= '0';
		end if;
  end if;
end process;
process(l_clk7x,locked)
begin
	if(locked = '0')then
		l_clk_tgl <= '0';
		l_clk_pq(1) <= '0';
		l_clk_nq(2 downto 1) <= "00";
		downcnt <= '0';
  elsif(l_clk7x'event and l_clk7x = '0')then
		l_clk_tgl <= l_clk_tgl xor (not l_clk_nq(2) and l_clk_nq(1));			
		l_clk_pq(1) <= l_clk_pq(0);
		l_clk_nq(2 downto 1) <= l_clk_nq(1 downto 0);
		downcnt <= (not l_clk_pq(1) and not l_clk_nq(2) and l_clk_nq(1)) or (downcnt and (l_clk_nq(2) or not l_clk_nq(1)));
  end if;
end process;
process(l_clk7x)
begin
  if(l_clk7x'event and l_clk7x = '0')then
--		l_clk_pq(1) <= l_clk_pq(0);
--		l_clk_nq(2 downto 1) <= l_clk_nq(1 downto 0);
--		downcnt <= (not l_clk_pq(1) and not l_clk_nq(2) and l_clk_nq(1)) or (downcnt and (l_clk_nq(2) or not l_clk_nq(1)));
  end if;
end process;
process(clk,locked)
begin
	if(locked = '0')then
		l_clk_sync <= "000";
		downcnt_sync <= '0';
  elsif(clk'event and clk = '1')then
	  l_clk_sync <= l_clk_sync(1 downto 0) & l_clk_tgl;
		downcnt_sync <= downcnt;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
		if(clk_chk = '1')then
		  l_clk_cntr <= "0000001";
		elsif(l_clk_sync(2) /= l_clk_sync(1))then
		  l_clk_cntr <= l_clk_cntr + 1;
		end if;
		if(clk_chk = '1')then
		  bound <= "00"; 
		else
		  if(l_clk_cntr(6 downto 3) = x"f")then
				bound(0) <= '1'; 
			end if;
		  if(bound(0) = '1' and l_clk_cntr(1 downto 0) = "11")then
				bound(1) <= '1'; 
			end if;
		end if;
		if(clk_chk = '1')then
			if(locked = '1' or lock_cntr(8) = '1')then
				lock_cntr <= (others => '0');
			else
				lock_cntr <= lock_cntr + 1;
			end if;
		end if;
		if(clk_chk = '1')then
			if(locked = '0')then
				err_cntr <= "00";
		  elsif(bound /= "01" and err_cntr /= "11")then
			  err_cntr <= err_cntr + 1;
			elsif(bound = "01" and err_cntr /= "00")then
			  err_cntr <= err_cntr - 1;
			end if;
		end if;
		if(err_cntr = "11" or lock_cntr(8) = '1')then
		  rst_dcm <= '1';
		elsif(err_cntr = "00")then
		  rst_dcm <= '0';
		end if;
  end if;
end process;
i_l_clk_q : IDDR2
   generic map(DDR_ALIGNMENT => "C1")
   port map (
      Q0 => l_clk_pq(0), -- 1-bit output captured with C0 clock
      Q1 => l_clk_nq(0), -- 1-bit output captured with C1 clock
      C0 => l_clk7x, -- 1-bit clock input
      C1 => not l_clk7x, -- 1-bit clock input
      CE => '1',  -- 1-bit clock enable input
      D => l_clk,   -- 1-bit data input 
      R => not locked,    -- 1-bit reset input
      S => '0'     -- 1-bit set input
   );
process(l_clk7x)
begin
  if(l_clk7x'event and l_clk7x = '1')then
	  l_data_q <= l_data;
		l_data_sr <= l_data_sr(23 downto 0) & (l_data_q xor polarity);
		l_clk_pq(4 downto 2) <= l_clk_pq(3 downto 1);
		strobe_i <= strobe_p; -- due to FPGA pin assignment, l_clk is inverted
		strobe_p <= l_clk_pq(3) and not l_clk_pq(4); -- one clk7x cycle earlier than strobe_i
		if(strobe_i = '1')then
		  q(0) <= l_data_sr(0);
		  q(1) <= l_data_sr(4);
		  q(2) <= l_data_sr(8);
		  q(3) <= l_data_sr(12);
		  q(4) <= l_data_sr(16);
		  q(5) <= l_data_sr(7);
		  q(6) <= l_data_sr(20);
		  q(7) <= l_data_sr(24);
		  q(8) <= l_data_sr(1);
		  q(9) <= l_data_sr(5);
		  q(10) <= l_data_sr(11);
		  q(11) <= l_data_sr(15);
		  q(12) <= l_data_sr(9);
		  q(13) <= l_data_sr(13);
		  q(14) <= l_data_sr(17);
		  q(15) <= l_data_sr(21);
		  q(16) <= l_data_sr(19);
		  q(17) <= l_data_sr(23);
		  q(18) <= l_data_sr(25);
		  q(19) <= l_data_sr(2);
		  q(20) <= l_data_sr(6);
		  q(21) <= l_data_sr(10);
		  q(22) <= l_data_sr(14);
		  q(23) <= l_data_sr(27);
		  q(24) <= l_data_sr(18);
		  q(25) <= l_data_sr(22);
		  q(26) <= l_data_sr(26);
		  q(27) <= l_data_sr(3);
		end if;
  end if;
end process;
end Behavioral;

